
#include <iostream>
#include <sstream>
#include <string>

#include <boost/shared_ptr.hpp>

#include "mlink/regiomlink.h"
#include "mlink/exc.h"
#include "mlink/Jtag_Impl_CCBPCIe.h"
#include "mlink/Jtag_Impl_Tracer.h"
#include "colorlog/log.h"
#include "svf/svfrun.h"

using boost::shared_ptr;
using namespace mlink;

int main(int argc, char *argv[]) {
    //  set_debug_level(5);

    std::string fileName;
    std::string address;
    if (argc == 3) {
        fileName = argv[1];
        address = argv[2];
    } else {
        std::cerr << "Usage: " << argv[0] << " <filename> <address>" << std::endl << std::endl;
        exit(EXIT_FAILURE);
    }
    std::istringstream istr(address);
    int addressNumeric = 0;
    istr >> addressNumeric;
    try {
        shared_ptr<Jtag_Impl> jtagImpl;
        shared_ptr<RegIOBase> link;
        try {
            link = shared_ptr<RegIOBase> (new RegIOMLink());
            link->reset();
            jtagImpl = shared_ptr<Jtag_Impl>(new Jtag_Impl_CCBPCIe(*link, addressNumeric));
        } catch (exc &e) {
            std::cerr << e.what() << std::endl;
            std::cout << "JTAG Tracer enabled" << std::endl;
            //	  jtagImpl = new Jtag_Impl_Tracer();
            jtagImpl = shared_ptr<Jtag_Impl>(new Jtag_Impl_Null());
        }
        //	jtagImpl->setTracing(false, true);
        SvfRun parser(jtagImpl, fileName);
        //	link->setLocalJtagMode(true);
        parser.exec();
        //	link->setLocalJtagMode(false);
        std::cout << "Success" << std::endl;
        //    delete jtagImpl;
    }
    catch (std::exception &e) {
        std::cerr << "Exception: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
