
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <getopt.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/poll.h>

#include <ctime>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cerrno>

#include "mlink/mlink.h"
#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/frame_hw.h"
#include "colorlog/log.h"
#include "commipc/commipc.h"

using namespace colorlog;

using namespace mlink;

//#include "ml_adcio.h"
//#include "ml_jtag.h"

//#include "adcm_io.h"

#define ML_SEQ_NEXT (-1)


// Static data buffer. if test_val_set == 0, use prbs.

#define MIN_DATA_WORDS ML_ADCM_MIN_FRAME_DATA_WORDS
#define MAX_DATA_WORDS ML_ADCM_MAX_FRAME_DATA_WORDS
//#define MAX_DATA_WORDS 251
//enum { maxpar = ML_MAX_FRAME_DATA_WORDS };
//enum { maxpar = ML_ADCM_MAX_FRAME_DATA_WORDS };

//int test_val[maxpar];
//int test_val_set = 0;

int extloop=0, sloop=0, loop_forever=0;
const char *testname = "unknown";

char devname[1024];
/*
enum { sendbuf_cap = ML_MAX_FRAME_DATA_WORDS*4};
enum { recvbuf_cap = ML_MAX_FRAME_DATA_WORDS*4};
unsigned int sendbuf[sendbuf_cap];
unsigned int recvbuf[recvbuf_cap];
*/

//static int c1=0;
static int cc = 0;
static int prbs_state = 1;
inline int prbs (int i)
{
    unsigned int r = i, k, bits = 0;
    cc++;
    // poly: x11+x13+x14+x16+1
    // shift left
    for (k = 0; k < 16; k++) {
        //  int val = 1 ^ ((r>>16)&1);
        int fb = 1 ^ ((r >> 15) & 1) ^ ((r >> 13) & 1) ^ ((r >> 12) & 1) ^ ((r >> 10) & 1);
        int out = r >> 15;          // bit#15
        r = ((r << 1) & 0xFFFE) | fb;
        //    r = (r<<1) | fb;
        //    printf("r=%04X, fb=%d\n", r, fb);
        bits |= (out ? (1 << k) : 0);
    }
    //  printf ("%d: %08X\n", cc, bits);
    //  if (cc==1000) c1=r; return(r);
    //  if (r==c1) printf("%08X repeat at %d\n", c1, cc);
    return r;
}

inline int prbs_next ()
{
    prbs_state = prbs (prbs_state);
    return prbs_state;
}
#if 0
void fill_buf (int *buf, int count)
{
    int n=0;
    if (test_val_set) {
        for (n = 0; n < count; n++) {
            buf[n] = test_val[n % test_val_set];
        }
    } else {
        for (n = 0; n < count; n++) {
            //	  buf[n] = prbs_next ();
            buf[n] = rand ();
            //  	  buf[n] = n | 0x5a000000;
        }
    }
    /*
  for (i = 0; i < count; i++) {
    int tt = 0;
    if (test_val_set) {
      tt = test_val[i % test_val_set];
    } else {
      tt = prbs_next ();
    }
//    buf[i] = tt;
//  buf[i]=0xFFFFFFFF;
        buf[i] = i | 0x5a000000;
  }
*/
}
#endif
// Benchmark timers
struct timeval tv1, tv2;
void timer_start ()
{
    ml_clear_rxtx_counters ();
    gettimeofday (&tv1, NULL);
}

void timer_stop ()
{
    gettimeofday (&tv2, NULL);
}

void timer_print ()
{
    int rxf = ml_get_rxf ();
    int rxb = ml_get_rxb ();
    int rxw = rxb / 4;
    double elapsed = (tv2.tv_sec - tv1.tv_sec) + 1e-6 * (tv2.tv_usec - tv1.tv_usec);
    if (elapsed == 0)
        elapsed = 1;
    log_info ("Time: %.2f s, %.1f fps, %.3f kwps", elapsed, 1. * rxf / elapsed, 1e-3 * rxw / elapsed);
}

void timer_stop_print ()
{
    timer_stop ();
    timer_print ();
}

/*****************************************************************************/

void banner ()
{
    fprintf (stderr, "M-Link Test Program\n\n");
}

void usage (const char *progname)
{
    printf ("\nUsage: %s\n\n", progname);
    printf ("   -h, --help              show this help\n");
    printf ("   -m, --dev <devname>     use device file <devname>\n");
    printf ("   -s, --sloop             SerDes/Mezzanine loopback test\n");
    printf ("   -x, --extloop           External (cable stub) loopback test\n");
    printf ("   -f, --forever           Repeat test continuously\n");
    //  printf ("   -r, --raw-dump[=NUM]    enable raw dump (show first NUM lines)\n");
    //  printf ("   -t, --dma-timeout=NUM   set dma timeout to NUM milliseconds (default %d)\n", dma_timeout_ms);
    //  printf ("   -f, --dma-fragment=NUM  set dma fragment size to NUM bytes (default %d)\n", dma_frag_size);
    //  printf ("   -b, --dma-size=NUM      set dma buffer size to NUM bytes (default %d)\n", dma_buf_size);
    //  printf ("   -k, --storage-keep      do not recycle storage data file every burst\n");
    //  printf ("   -a, --auto              automatic mode (hardware constants are loaded default\n");
    //  printf ("   -n, --notest            disable system selftest at startup\n");
    printf ("   -d, --debug[=LEVEL]     set debug level to LEVEL, or increase by 1 if not specified\n");
    printf ("\n");
}

void parse_cmdline (int argc, char *const argv[])
{
    int c;
    int dl = 0;
    int go_exit = 0, show_help = 0;
    char str[20];
    long d;
    char *endptr;
    int option_index = 0;

    static struct option long_options[] = {
        {"debug", optional_argument, 0, 'd'},
        {"help", no_argument, 0, 'h'},
        {"extloop", no_argument, 0, 'x'},
        {"sloop", no_argument, 0, 's'},
        {"devname", required_argument, 0, 'm'},
        {"forever", no_argument, 0, 'f'},
        //    {"raw-dump", optional_argument, 0, 'r'},
        //    {"dma-timeout", required_argument, NULL, 't'},
        //    {"dma-fragment", required_argument, NULL, 'f'},
        //    {"dma-size", required_argument, NULL, 'b'},
        //    {"stor-keep", no_argument, 0, 'k'},
        //    {"auto", no_argument, 0, 'a'},
        //    {"notest", no_argument, 0, 'n'},
        {NULL, 0, NULL, 0}
    };
    strcpy (str, "d::");
    //  strcat (str, "hr::t:f:b:kan");
    strcat (str, "hxsm:b:f");
    while (1) {
        c = getopt_long (argc, argv, str, long_options, &option_index);
        if (c == -1)
            break;
        /*
   if ((c != ':')&&(c != '?')) {
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
  }
*/
        switch (c) {
        case 'd':
            if (optarg) {
                d = strtol (optarg, &endptr, 10);
                if ((*endptr == 0) && (d >= 0)) {
                    dl = d;
                } else {
                    fprintf (stderr, "bad option 'd' value: %s\n", optarg);
                }
            } else {
                dl++;
            }
            break;
        case 'x':
            extloop=1;
            break;
        case 's':
            sloop=1;
            break;
        case 'f':
            loop_forever=1;
            break;
        case 'm':
            if (optarg) {
                strncpy(devname, optarg, sizeof(devname));
                devname[sizeof(devname)-1] = 0;
            }
            break;
        case 'h':
            show_help = 1;
            go_exit = 1;
            break;
            /*
    case 'r':
      set_raw_dump = raw_burst_dump_maxlines;
      if (optarg) {
        d = strtol (optarg, &endptr, 10);
        if ((*endptr == 0) && (d > 0)) {
          set_raw_dump = d;
        } else {
          fprintf (stderr, "bad option '%c' value: %s\n", c, optarg);
        }
      }
      break;

*/
        case '?':
            show_help = 1;
            go_exit = 2;
            break;
        default:
            fprintf (stderr, "?? getopt returned character %c (code 0%o) ??\n", c, c);
            show_help = 1;
            go_exit = 2;
            break;
        }
    }
    /*
  while (optind < argc) {
    char *s = argv[optind];
    int n;
    if (strlen (s) > 2 && '0' == s[0] && 'x' == s[1])
      n = sscanf (&s[2], "%08X", &test_val[test_val_set]);
    else
      n = sscanf (s, "%d", &test_val[test_val_set]);
    if (n != 1) {
      fprintf (stderr, "Bad parameter: %s\n", s);
      show_help = 1;
      go_exit = 2;
      break;
    }
    test_val_set++;
    optind++;
  }
*/
    if (show_help) {
        usage (argv[0]);
        if (!go_exit)
            go_exit = 1;
        exit (go_exit - 1);
    }

    if (go_exit)
        exit (go_exit - 1);

    if (dl) {
        //    printf ("debug level: %d\n", dl);
        set_debug_level (dl);
    }
    if (go_exit)
        exit (go_exit - 1);
}

/*****************************************************************************/

//const char rotbits[16] = { 15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0};
const char rotbits[16] = { 0,8,4,12,2,10,6,14,1,9,5,13,3,11,7,15};

unsigned int rotw(unsigned int x)
{
    int i;
    unsigned int tmp=0;
    for(i=0;i<8;i++)
        tmp |= rotbits[(x>>(4*i)) & 0xF] << (4*(7-i));
    return tmp;
}

int main (int argc, char *const argv[])
{
    int ret = -1, test_ok = 1;
    mlink_caps_t devcaps; // device capabilities
    //  int bad_usage = 0;

    strcpy(devname, ""); // use default
    banner ();
    parse_cmdline (argc, argv);

    //  fill_buf (sendbuf, sizeof(sendbuf)/4);


    /*
  if (argc > maxpar + 1)
    bad_usage = 1;
  {
    int i;
    for (i = 0; (i < (argc - 1)) && !bad_usage; i++) {
      const char *s = argv[i + 1];
      int n = 0;
      if (strlen (s) > 2 && '0' == s[0] && 'x' == s[1])
        n = sscanf (&s[2], "%08X", &test_val[test_val_set]);
      else
        n = sscanf (s, "%d", &test_val[test_val_set]);
      if (n != 1) {
        fprintf (stderr, "Bad parameter: %s\n", s);
        bad_usage = 1;
        break;
      }
// printf("test [%d]: %08X\n", i, test_val[i]);
      test_val_set++;
    }
    if (bad_usage) {
      fprintf (stderr, "USAGE: %s [number]\n", argv[0]);
      exit (EXIT_FAILURE);
    }
  }
*/
    //  init_log ("fds", LOG_PID | LOG_CONS | LOG_NDELAY, LOG_LOCAL6);
    //  set_debug_level (3);

    if ((ret = ml_open (devname))) {
        log_err ("Failed to open interface - no device found?");
        return (ret);
    }
    devcaps.all = 0;
    ml_get_caps (&devcaps);

    if (sloop && 0==devcaps.bit.serdes_loop_test) {
        log_warn("Device does not support serdes loopback test");
        return(1);
    }
    /*
// No need to allocate DMA buffer for simple tests - but it is still required. FIXME
  if ((ret = ml_init_dma_buffer (ML_DMABUF_RDWR, 16000 * 4096))) {
    log_err ("Failed to allocate dma buffer");
    return (ret);
  }
  log_debug (1, "dma buffer allocated");
*/
    /*
//  sleep(5);
  if ((ret = ml_free_dma_buffer ())) {
    log_err ("Failed to free dma buffer");
        return(ret);
  }
  log_debug(1, "dma buffer freed");

  if ((ret = ml_close ())) {
    log_err ("Failed to close interface");
        return(ret);
  }
  log_debug(1, "device closed");
return(0);
*/

    // Internal loopback (CCB-PCI)
    if ((ret = ml_reset ())) {
        log_err ("Failed to reset interface");
        return (ret);
    }
    log_info ("Reset complete");

    // For external loopback make sure Rx and Tx paths are connected by stub.
    if (sloop || extloop) {
        int driver_loop_mode = 0;
        if (sloop) { testname = "SerDes Loopback"; driver_loop_mode = 1; }
        if (extloop) { testname = "Remote Loopback"; driver_loop_mode = 2; }
        if ((ret = ml_set_ext_loop_mode (driver_loop_mode))) {
            log_err ("Failed to set %s test mode", testname);
            return (ret);
        }
        usleep(1000);
        log_debug(1, "%s test...", testname);
        while(1) {
            int need_reset = 0;
            timer_start ();
            //      ret = ml_loopback_test (10000, 332, 0);
            ret = ml_loopback_test (10000, ML_MAX_FRAME_DATA_WORDS, 0);
            timer_stop_print ();
            if (ret) {
                log_err ("%s test FAILED", testname);
                test_ok = 0;
                need_reset = 1;
            } else {
                log_info ("%s test OK", testname);
            }
            if (need_reset) {
                /*
                if ((ret = ml_reset ())) {
                  log_err ("Failed to reset interface");
                  return (ret);
                }
*/
                //		break; // interrupt test on error
                usleep(10000);
                if (!loop_forever) break; // interrupt test on error
            }
        }
    }

    //  sleep(1000);
    /*
  if ((ret = ml_free_dma_buffer ())) {
    log_err ("Failed to free dma buffer");
  } else {
    log_debug (1, "dma buffer freed");
  }
*/
    if ((ret = ml_close ())) {
        log_err ("Failed to close interface");
    } else {
        log_debug (1, "device closed");
    }
    if (test_ok) {
        log_info("Test Ok");
        return (0);
    } else {
        log_warn("Test FAILED");
        return (1);
    }
}
