
#include "BitVect.h"

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <string>
#include <sstream>
#include <iostream>
#include <stdexcept>

#include "mlink/stl_ext.h"

//using namespace std;

namespace mlink {

    std::ostream &operator<< (std::ostream & os, const std::vector<bool> &r)
    {
        os << "{";
        for(std::vector<bool>::const_iterator it = r.begin(); it != r.end(); ++it) {
            os << (*it);
        }
        os << "}";
        return os;
    }

    std::ostream &operator<< (std::ostream & os, const std::vector<int> &r)
    {
        os << "{";
        if (!r.empty()) {
            os << r[0];
            for(std::vector<int>::const_iterator it = ++r.begin(); it!=r.end(); ++it) {
                os << ", " << (*it);
            }
        }
        os << "}";
        return os;
    }



    int charHexBin(char c)
    {
        char s[2];
        s[0] = c;
        s[1] = '\0';
        char *endptr=0;
        long int val = strtol(s, &endptr, 16);
        if (errno) {
            std::ostringstream os;
            os << "charHexBin: " << strerror(errno);
            throw std::runtime_error(os.str());
        }
        if (*endptr != '\0') {
            std::ostringstream os;
            os << "charHexBin: bad input '" << c << "' (code " << (unsigned int)c << ")";
            throw std::runtime_error(os.str());
        }
        return val;
    }

    BitVect::BitVect(size_t bitCount, const std::string &s)
    {
        //  cout << "Converting string '" << s << "'" << endl;
        size_t charCount = (bitCount+3)/4;
        if (s.size() < charCount) {
            std::ostringstream ost;
            ost << "insufficient hex string length: " << s.size() << ", expected " << charCount << std::endl;
            throw std::runtime_error(ost.str());
        }
        resize(charCount*4); // align by 4
        size_t bits = 0;

        for(std::string::const_reverse_iterator rit = s.rbegin(); rit!=s.rend(); rit++) {
            const char c = (*rit);
            int d = charHexBin(c);
            (*this)[bits] = d & 0x1;
            (*this)[bits+1] = d & 0x2;
            (*this)[bits+2] = d & 0x4;
            (*this)[bits+3] = d & 0x8;
            //	cout << "char=" << c << ", bits=" << bits << ", d=" << d << endl;
            bits += 4;
        }
        resize(bitCount); // trim
    }

    void BitVect::push_back(const BitVect &v)
    {
        insert(end(), v.begin(), v.end());
    }
    /*
void BitVect::push_front(const BitVect &v)
{
  insert(begin(), v.begin(), v.end());
}
*/
    /*
std::ostream &operator<< (std::ostream & os, const BitVect &r)
{
  os << "BitVect[" << r.size() << "], bits={";
  for(BitVect::const_reverse_iterator it = r.rbegin(); it!=r.rend(); it++) {
        os << (*it);
  }
  os << "}";
  return os;
}
*/
    void BitVect::print() const
    {
        BitVect v(*this);
        std::cout << v;
        std::cout << std::endl;
    }

    /*
std::string &BitVect::operator<< (std::string & out)
{
  ostringstream ost;
  ost << (*this);
  out.append(ost.str());
  return out;
}
*/
    void BitVect::test()
    {
        testOperatorPlus();
        testOperatorAnd();
    }
    void BitVect::testOperatorPlus()
    {
        {
            const BitVect v1(11, "FFF"); // 11111111111
            const BitVect v2(8, "55"); // 01010101
            BitVect v3 = v1;
            v3.push_back(v2);
            BitVect vt(19, "2AFFF"); // 01010101 11111111111
            if (!(vt == v3)) {
                std::cerr << "FAILED BitVect::testOperatorPlus() 1" << std::endl;
                v1.print();
                v2.print();
                v3.print();
                vt.print();
                throw std::exception();
            }
            /*
  v3 = v1;
  v3.push_front(v2);
  vt = BitVect(19, "7FF55"); // 11111111111 01010101
  if (!(vt == v3)) {
        cerr << "FAILED BitVect::testOperatorPlus() 2" << endl;
        v1.print();
        v2.print();
        v3.print();
        vt.print();
        throw exception();
  }
*/
        }
        /*
 {
  const BitVect v1(11, "153"); // .001 0101 0011
  const BitVect v2(3,   "01"); // .... .001
  BitVect v3 = v1;
  v3.push_front(v2);
  BitVect vt(14, "0A99"); // 00101010011 001
  if (!(vt == v3)) {
        cerr << "FAILED BitVect::testOperatorPlus() 3" << endl;
        v1.print();
        v2.print();
        v3.print();
        vt.print();
        throw exception();
  }
 }
*/
    }

    void BitVect::testOperatorAnd()
    {
        {
            const BitVect v1(32, "921FCDF4");
            const BitVect v2(32, "067B3559");
            BitVect v3 = v1 & v2;
            BitVect vt(32, "0x21B0550");
            if (!(vt == v3)) {
                std::cerr << "FAILED BitVect::testOperatorAnd() 1" << std::endl;
                v1.print();
                v2.print();
                v3.print();
                vt.print();
                throw std::exception();
            }
        }
    }

    BitVect BitVect::operator&(const BitVect &v2) const
    {
        if (size() != v2.size())
            throw std::runtime_error("size mismatch");
        BitVect v(*this);
        for(size_t i=0; i<size(); i++)
            v[i] = v[i] & v2[i];
        return v;
    }

};
