
#include "JtagPath.h"

#include <iostream>
#include <sstream>
#include <stdexcept>
#include <algorithm>

#include "mlink/stl_ext.h"

namespace mlink {

    /*
std::ostream &operator<< (std::ostream & os, const JtagPath &r)
{
  os << "JtagPath[" << r.size() << "]";
  os << ", tms={";
  foreach_const(r, it) {
        os << (*it);
  }
  os << "}";

  return os;
}
*/
    JtagPath::~JtagPath()
    {
        //  cout << (*this) << endl;
    }

    struct PathStruct {
        JtagStatePath statePath;
        JtagTmsPath tmsPath;
    };

    std::ostream &operator<< (std::ostream & os, const JtagStatePath &r)
    {
        os << "{";
        if (! r.empty()) {
            os << r[0];
            for(JtagStatePath::const_iterator it = ++r.begin(); it!=r.end(); ++it) {
                os << ", " << (*it);
            }
        }
        os << "}";
        return os;
    }

    PathStruct findState(const PathStruct &ps, const JTAG_STATE &finish)
    {
        //  cout << ps.statePath << endl;
        if (ps.statePath.empty()) {
            throw std::runtime_error("Starting point not defined");
        }
        if (ps.statePath.size() > JTAG_STATE_COUNT) {
            throw std::runtime_error("Path recursion limit reached in findState");
        }
        const JTAG_STATE &start = ps.statePath[ps.statePath.size()-1];
        const JTAG_STATE state0 = jtag_next_state[start][0];
        const JTAG_STATE state1 = jtag_next_state[start][1];
        //  cout << ps.statePath << " next tries: " << state0 << ", " << state1 << endl;
        PathStruct p0(ps); p0.statePath.push_back(state0); p0.tmsPath.push_back(0);
        PathStruct p1(ps); p1.statePath.push_back(state1); p1.tmsPath.push_back(1);
        if (state0 == finish) { return p0; }
        if (state1 == finish) { return p1; }
        // check for loop: if next point is not walked through already

        bool loop0 = find(ps.statePath.begin(), ps.statePath.end(), state0) != ps.statePath.end();
        bool loop1 = find(ps.statePath.begin(), ps.statePath.end(), state1) != ps.statePath.end();

        // Empty == invalid
        PathStruct newPath0;
        PathStruct newPath1;
        if (!loop0) newPath0 = findState(p0, finish);
        if (!loop1) newPath1 = findState(p1, finish);

        const size_t s0 = newPath0.statePath.size();
        const size_t s1 = newPath1.statePath.size();

        if (s0 > 0 && s1 > 0) {
            //	cout << "Comparing paths:" << endl;
            //	cout << newPath0.statePath << endl;
            //	cout << newPath1.statePath << endl;
            if (s0 < s1) {
                //	  cout << "Shortest path is:" << endl << newPath0.statePath << endl << endl;
                //	  cout << "Shortest path is " << newPath0.tmsPath << endl;
                return newPath0;
            } else {
                //	  cout << "Shortest path is:" << endl << newPath1.statePath << endl << endl;
                //	  cout << "Shortest path is " << newPath1.tmsPath << endl;
                return newPath1;
            }
        }
        if (s0) {
            //	cout << "Single path:" << endl << newPath0.statePath << endl << endl;
            return newPath0;
        }
        if (s1) {
            //	cout << "Single path:" << endl << newPath1.statePath << endl << endl;
            return newPath1;
        }
        //  cout << "Both paths terminated" << endl;
        return PathStruct();
    }

    JtagPath::JtagPath(const JTAG_STATE &start, const JTAG_STATE &finish)
    {
        startState = start;
        finishState = finish;
        PathStruct path;
        // Special case: TEST_LOGIC_RESET requires 5 TMS, assume start is unknown
        if (finishState == JS_TEST_LOGIC_RESET) {
            path.statePath = JtagStatePath(6, JS_TEST_LOGIC_RESET);
            path.tmsPath = std::vector<int>(5, 1);
        } else {
            //  cout << "------------------------------------------------------------------------" << endl;
            //  cout << "Searching path: " << startState << " -> " << finishState << endl;
            PathStruct p;
            p.statePath = JtagStatePath(1, startState);
            path = findState(p, finishState);
        }
        statePath = path.statePath;
        tmsPath = path.tmsPath;
        if (statePath.empty() || statePath[statePath.size()-1] != finishState) {
            throw std::runtime_error("FAILED");
        }
        //  cout << "Result: " << path << endl;
        //  cout << "TMS sequence: " << tms << endl;
        //  cout << "Result: " << startState << " " << finishState << " " << statePath << " " << tmsPath << endl;
        //  cout << "JtagPath[" << startState << "][" << finishState << "] = " << tmsPath << ";" << endl;
        //  cout << endl;
    }

};
