
#include "svfrun.h"

#include <cstdlib>
#include <deque>
#include <iostream>
#include <sstream>
#include <stdexcept>

#include "mlink/jtag_state.h"
#include "mlink/jtag_impl.h"
#include "JtagPath.h"
#include "BitVect.h"
#include "mlink/stl_ext.h"
#include "usleep.h"

//#include <boost/numeric/ublas/vector.hpp>

namespace mlink {

    //using namespace std;
    //using namespace boost::numeric::ublas;

    void SvfRun::exec(bool enableCheckException)
    {
        _enableCheckException = enableCheckException;
        //  cout << endl << "Exec:" << endl;
        JTAG_STATE tapState = JS_RUN_TEST_IDLE;
        for(SvfList::const_iterator it = commands.begin(); it != commands.end(); ++it) {
            //	const SvfOp *cmd = it;
            const JtagShift jtagShift = (*it)->getSequence();
            double delay = (*it)->getDelay();
            const std::string &comment = (*it)->getComment();
            if (!comment.empty()) {
                std::cout << comment << std::endl;
            }
            const JTAG_STATE endState = (*it)->endState();
            if (!jtagShift.empty()) {
                const JTAG_STATE beginState = (*it)->beginState();
                if (beginState != tapState) {
                    runStateChange(tapState, beginState);
                    tapState = beginState;
                }
                runShift(jtagShift);
                // shift moves to next state only if endState is not shift_xr
                if (endState != tapState) {
                    tapState = jtag_next_state[tapState][1];
                }
            }
            if (endState != tapState) {
                runStateChange(tapState, endState);
                tapState = endState;
            }
            if (delay>0) {
                std::cout << "Delay " << delay << std::endl;
                sleep(static_cast<unsigned long>(delay));
            }
            //	cout << " begin " << beginState << ", end " << endState << ", jtag " << jtagShift << endl;
            //    cout << endl;
        }
        //  cout << endl;
    }

    void SvfRun::runStateChange(const JTAG_STATE &start, const JTAG_STATE &finish)
    {
        //  std::cout << start << " -> " << finish << std::endl;
        JtagTmsPath tmsPath = JtagPath(start, finish).getTmsPath();
        for(JtagTmsPath::const_iterator it = tmsPath.begin(); it != tmsPath.end(); ++it) {
            bool tms = 0 != (*it);
            bool tdi = 0;
            bool tdo = jtagImpl->jtag_io(tms, tdi);
            tdo = tdo; // unused
            //	cout << "TMS: " << tms << ", TDI: " << tdi << ", TDO: " << tdo << endl;
        }
    }

    void SvfRun::runShift(const JtagShift &jtagShift)
    {
        bool tdoCompare = true;
        //  std::cout << jtagShift << std::endl;
        if (jtagShift.empty()) return;
        /*
  foreach_const(jtagShift, it) {
        const JtagStep &op = *it;
        bool tdo = jtagImpl->jtag_io(op.tms, op.tdi);
        vtdo.push_back(tdo);
        vtest.push_back(op.tdo);
        if ((op.mask & tdo) != (op.mask&op.tdo)) {
          tdoCompare = false;
        }
  }jtagShift
*/
        const size_t n = jtagShift.size();
        std::vector<int> vtest(n), vtdo(n), vmask(n);
        std::vector<bool> btms(n);
        std::vector<bool> btdi(n);
        for(size_t i=0; i<n; i++) {
            const JtagStep &op = jtagShift[i];
            btms[i] = op.tms;
            btdi[i] = op.tdi;
            // op.smask not used
            vtest[i] = op.tdo;
            vmask[i] = op.mask;
        }
        const std::vector<bool> &btdo = jtagImpl->jtag_io_b(btms, btdi);


        for(size_t i=0; i<n; i++) {
            vtdo[i] = btdo[i];
        }
        for(size_t i=0; i<n; i++) {
            //	vtdo[i] = vtdo[i] & vmask[i];
            //	vtest[i] = vtest[i] & vmask[i];
            if ((vtdo[i]&vmask[i]) != (vtest[i]&vmask[i])) {
                tdoCompare = false;
                break;
            }
        }

        if (!tdoCompare) {
            std::ostringstream os;
            os << "TDO expected: " << std::endl << vtest << std::endl
                    << "TDO received: " << std::endl << vtdo  << std::endl
                    << "TDO mask:     " << std::endl << vmask << std::endl;
            std::cout << os.str();
            if (_enableCheckException) {
                throw std::runtime_error(os.str());
            }
        }
    }

};
